"""
Base class abstraction for an ingestor. An Ingestor is responsible for taking a
document (or multiple documents) and uploading them to a given RAG service.
"""
# Standard
from abc import abstractmethod

# First Party
import aconfig

# Local
from ..factory import FactoryConstructible
from ..storage import StorageBase
from ..types import Document


class Ingestor(FactoryConstructible):
    __doc__ = __doc__

    @abstractmethod
    def __init__(
        self,
        config: aconfig.Config,
        instance_name: str,
        *,
        storage: StorageBase,
    ):
        """Construct with a storage instance and the factory config"""

    @abstractmethod
    def ingest(self, documents: list[Document]):
        """Ingest a document or a list of documents

        Args:
            documents: list of documents to ingest
        """

    @abstractmethod
    def delete(self, documents: list[Document]):
        """Delete the set of documents from the RAG instance"""



"""
Unit tests for testing the AnythingLLM ingestor
"""
# Standard
from contextlib import contextmanager
from unittest import mock
import io
import json
import os
import re
import uuid

# Third Party
import pytest
import requests

# First Party
import aconfig

# Local
from ragnardoc.ingestors.anything_llm import AnythingLLMIngestor
from ragnardoc.storage import storage_factory
from ragnardoc.types import Document
from tests.conftest import ServerMockBase

## Helpers #####################################################################


class AnythingLLMMock(ServerMockBase):
    """Mock implementation of AnythingLLM that stores docs in memory"""

    def __init__(self, base_url: str, workspaces: list[str]):
        self.base_url = base_url
        self.docs = {"custom-documents": {}}
        self.workspaces = {
            self._workspace_slug(ws): {
                "name": ws,
                "slug": self._workspace_slug(ws),
                "documents": [],
            }
            for ws in workspaces
        }

    def _handle_call(self, method: str, url: str, body: dict | None = None):
        """Handle an API call"""
        if method == "post":
            if url == f"{self.base_url}/api/v1/document/raw-text":
                return self._upload(body)
            if url == f"{self.base_url}/api/v1/document/move-files":
                return self._move(body)
            if url == f"{self.base_url}/api/v1/document/create-folder":
                return self._create_folder(body)
            if (workspace_slug := self._get_workspace_slug(url)) and url.endswith(
                f"{workspace_slug}/update-embeddings"
            ):
                return self._update_embeddings(workspace_slug, body)
        elif method == "get":
            if url == f"{self.base_url}/api/v1/workspaces":
                return self._get_workspaces()
            if (workspace_slug := self._get_workspace_slug(url)) and url.endswith(
                workspace_slug
            ):
                return self._workspace_details(workspace_slug)
        elif method == "delete":
            if url == f"{self.base_url}/api/v1/system/remove-documents":
                return self._delete_documents(body)
        return self._make_error(404, "Not found")

    @classmethod
    def _make_error(cls, status_code: int, msg: str = "") -> requests.Response:
        return cls._make_response({"success": False, "error": msg}, status_code)

    def _folder_exists(self, folder_name: str) -> bool:
        return folder_name in self.docs

    @staticmethod
    def _workspace_slug(name: str) -> str:
        return name.replace(".", "-dot-").replace(":", "")

    def _get_workspace_slug(self, url: str) -> str | None:
        if m := re.match(f"{self.base_url}/api/v1/workspace/([^/]+)", url):
            return m.group(1)

    def _split_folder_name(self, docpath: str) -> tuple[str | None, str | None]:
        parts = docpath.split(os.sep)
        if len(parts) != 2 or not self._folder_exists(parts[0]):
            return None, None
        return parts

    ## Handlers ##

    def _upload(self, body: dict) -> requests.Response:
        """Upload always places the doc in custom-documents"""
        try:
            title = body["metadata"]["title"]
        except IndexError:
            return self._make_error(422, "missing metadata.title")
        uid = str(uuid.uuid4())
        folder = "custom-documents"
        doc_path = f"raw-{title}-{uid}.json"
        self.docs[folder][doc_path] = body
        return self._make_response(
            {
                "success": True,
                "error": None,
                "documents": [
                    {
                        "id": uid,
                        "title": title,
                        "location": os.path.join(folder, doc_path),
                    }
                ],
            }
        )

    def _move(self, body: dict) -> requests.Response:
        """Move the location key"""
        for file_move in body.get("files", []):
            try:
                from_loc = file_move["from"]
                to_loc = file_move["to"]
                from_folder, from_name = self._split_folder_name(from_loc)
                to_folder, to_name = self._split_folder_name(to_loc)
                if None in [from_folder, from_name, to_folder, to_name]:
                    return self._make_error(500, "Failed to move some files")
                self.docs[to_folder][to_name] = self.docs[from_folder].pop(from_name)
            except KeyError:
                return self._make_error(500, "missing to/from")
        return self._make_response({"success": True})

    def _create_folder(self, body: dict) -> requests.Response:
        """Create or error if folder exists already to mirror the real server"""
        try:
            name = body["name"]
        except KeyError:
            return self._make_error(500, "Invalid path")
        if self._folder_exists(name):
            # NOTE: This specific error message is checked as OK
            return self._make_error(500, "Folder by that name already exists")
        self.docs[name] = {}
        return self._make_response({"success": True})

    def _get_workspaces(self) -> requests.Response:
        """Get details (without docs) for all workspaces"""
        return self._make_response(
            {
                "workspaces": [
                    {"name": details["name"], "slug": details["slug"]}
                    for details in self.workspaces.values()
                ]
            }
        )

    def _workspace_details_impl(self, workspace_slug: str) -> dict:
        """Get the detailed view of the workspace"""
        if workspace_slug not in self.workspaces:
            # 200 on bad workspace is what the real server does
            return self._make_response({"workspace": []})
        ws_details = self.workspaces[workspace_slug]
        return {"workspace": [ws_details]}

    def _workspace_details(self, workspace_slug: str) -> requests.Response:
        return self._make_response(self._workspace_details_impl(workspace_slug))

    def _update_embeddings(self, workspace_slug: str, body: dict) -> requests.Response:
        """Update the docs for the given workspace"""
        ws_details = self.workspaces.get(workspace_slug)
        if ws_details is None:
            return self._make_response("Bad Request", 400)
        ws_docs = ws_details.setdefault("documents", [])
        ws_doc_paths = [doc["docpath"] for doc in ws_docs]
        for add_path in body.get("adds", []):
            if add_path not in ws_doc_paths and None not in self._split_folder_name(
                add_path
            ):
                ws_docs.append({"docpath": add_path})
        for delete_path in body.get("deletes", []):
            if delete_path in ws_doc_paths:
                delete_idxs = [
                    idx
                    for idx, doc in enumerate(ws_docs)
                    if doc["docpath"] == delete_path
                ]
                assert len(delete_idxs) == 1, "Not sure what happens here!!!"
                ws_docs.pop(delete_idxs[0])
        ws_details = self._workspace_details_impl(workspace_slug)
        # It's a dict here and a list in the GET call
        return self._make_response({"workspace": ws_details["workspace"][0]})

    def _delete_documents(self, body: dict) -> requests.Response:
        """Delete docs if they've been unlinked from all workspaces"""
        for doc_path in body.get("names", []):
            folder, name = self._split_folder_name(doc_path)
            if None not in [folder, name]:
                self.docs[folder].pop(name, None)
        return self._make_response(
            {"success": True, "message": "Documents removed successfully"}
        )


@contextmanager
def anythingllm_mock_ctx(workspaces: list[str]):
    base_url = "http://localhost:5432187"
    mock_server = AnythingLLMMock(base_url, workspaces)
    with (
        mock.patch("requests.get", mock_server.get),
        mock.patch("requests.post", mock_server.post),
        mock.patch("requests.delete", mock_server.delete),
    ):
        yield mock_server


@pytest.fixture
def anythingllm():
    workspaces = ["workspace1", "workspace2"]
    with anythingllm_mock_ctx(workspaces) as mock_server:
        storage = storage_factory.construct({"type": "dict"})
        cfg = aconfig.Config(
            {
                "base_url": mock_server.base_url,
                "workspaces": workspaces,
                "apikey": "my-key",
                "root_folder": "ragnardoc_tests",
            },
            override_env_vars=False,
        )
        inst = AnythingLLMIngestor(cfg, "test-inst", storage=storage)
        inst.mock = mock_server
        yield inst


## Tests #######################################################################


def test_anythingllm_get_workspaces(anythingllm):
    """Test that the workspaces are fetched correctly"""
    assert set(anythingllm._workspace_slugs) == {"workspace1", "workspace2"}


def test_anythingllm_ingest_delete(anythingllm, data_dir):
    """Test basic ingestion and deletion"""
    docs = [
        Document.from_file(data_dir / "sample.txt", data_dir),
        Document.from_file(data_dir / "sample_docs" / "README.md", data_dir),
    ]
    anythingllm.ingest(docs)

    # Make sure root folder set up correctly
    assert anythingllm._root_folder in anythingllm.mock.docs

    # Make sure both docs found in correct folder
    assert len(anythingllm.mock.docs[anythingllm._root_folder]) == 2

    # Make sure docs are associated with both workspaces
    assert len(anythingllm.mock.workspaces["workspace1"]["documents"]) == 2
    assert len(anythingllm.mock.workspaces["workspace2"]["documents"]) == 2

    # Delete one of the docs
    delete_docs = [docs[0]]
    anythingllm.delete(delete_docs)

    # Make sure only one doc is left
    assert len(anythingllm.mock.docs[anythingllm._root_folder]) == 1

    # Make sure doc removed from both workspaces
    assert len(anythingllm.mock.workspaces["workspace1"]["documents"]) == 1
    assert len(anythingllm.mock.workspaces["workspace2"]["documents"]) == 1



"""
The start command initializes ragnardoc to run as a service that continuously
maintains the state of your documents in all of your connected RAG apps.
"""
# Standard
from datetime import timedelta
import argparse
import re
import shlex
import subprocess
import sys
import time

# First Party
import alog

# Local
from .. import config
from .base import CommandBase

log = alog.use_channel("START")


class StartCommand(CommandBase):
    __doc__ = __doc__
    name = "start"

    def __init__(self):
        self._period = self._parse_time(config.service.period)
        self._cmd = f"{sys.executable} -m ragnardoc run"
        self._running = False

    def add_args(self, parser: argparse.ArgumentParser):
        """Add the args to configure the periodic scraping"""
        parser.add_argument(
            "--period",
            "-p",
            default=None,
            help="The period to run the ingestion service",
        )

    def stop(self):
        self._running = False

    def run(self, args: argparse.Namespace):
        """Start the infinite loop to run periodically"""
        period = self._period
        if args.period:
            period = self._parse_time(args.period)
        self._running = True
        while self._running:
            log.info("Running ingestion service")
            self._ingest()
            log.info("Sleeping for %s", period)
            time.sleep(period.total_seconds())

    def _ingest(self):
        """Run the ingestion as a subprocess. This is done so that config
        changes are re-parsed on very run.
        """
        with alog.ContextTimer(log.debug, "Ingestion done in: "):
            subprocess.run(shlex.split(self._cmd))

    @staticmethod
    def _parse_time(time_str: str) -> timedelta:
        """Parse a time string into a timedelta object"""
        pattern = r"(\d+\.?\d*)([dhms])\s*"
        seconds = 0
        for match in re.finditer(pattern, time_str):
            value = float(match.group(1))
            unit = match.group(2)
            if unit == "s":
                seconds += value
            elif unit == "m":
                seconds += value * 60
            elif unit == "h":
                seconds += value * 60 * 60
            elif unit == "d":
                seconds += value * 60 * 60 * 24
        if not seconds:
            raise ValueError(f"Invalid time string: {time_str}")
        return timedelta(seconds=seconds)



"""
Shared test setup and fixtures
"""
# Standard
from pathlib import Path
import abc
import io
import json
import os
import shutil
import tempfile

# Third Party
import pytest
import requests

# First Party
import alog

alog.configure(
    default_level=os.getenv("LOG_LEVEL", "error"),
    filters=os.getenv("LOG_FILTERS", ""),
    formatter="json" if os.getenv("LOG_JSON", "").lower() == "true" else "pretty",
    thread_id=os.getenv("LOG_THREAD_IDF", "").lower() == "true",
)


@pytest.fixture
def data_dir():
    return Path(__file__).parent / "data"


@pytest.fixture
def mutable_data_dir(data_dir):
    """Get a version of the data dir that the test can mutate without side
    effects
    """
    with tempfile.TemporaryDirectory() as working_dir:
        mutable_dir = Path(working_dir) / "data"
        shutil.copytree(data_dir, mutable_dir)
        yield mutable_dir


@pytest.fixture
def txt_data_file(data_dir):
    return data_dir / "sample.txt"


@pytest.fixture
def scratch_dir():
    with tempfile.TemporaryDirectory() as dirname:
        yield Path(dirname)


# Force RAGNARDOC_HOME to be a temporary directory that will be auto-cleaned up.
# This is done while importing conftest.py to avoid the import-time config
# parsing where user config is merged.
_tempdir = tempfile.TemporaryDirectory(suffix="ragnardoc")
os.environ["RAGNARDOC_HOME"] = _tempdir.name


@pytest.fixture(autouse=True)
def ignore_user_config():
    with tempfile.TemporaryDirectory() as temp_home:
        _tempdir.cleanup()
        os.environ["RAGNARDOC_HOME"] = temp_home
        yield
        _tempdir.cleanup()


## Server Mock Structure #######################################################


class ServerMockBase(abc.ABC):
    """Shared base class for mock server implementations"""

    ## Abstract ##

    @abc.abstractmethod
    def _handle_call(self, method: str, url: str, body: dict | None = None):
        """Server-specific router method"""

    ## Public ##

    def get(self, url, *_, **__) -> requests.Response:
        return self._handle_call("get", url)

    def post(self, url, json=None, files=None, *_, **__) -> requests.Response:
        if json and files:
            raise ValueError("Cannot specify both `json` and `files`.")
        body = json or files
        return self._handle_call("post", url, body)

    def delete(self, url, json=None, *_, **__) -> requests.Response:
        return self._handle_call("delete", url, json)

    ## Protected ##

    @staticmethod
    def _make_response(
        resp_body: dict | str, status_code: int = 200
    ) -> requests.Response:
        resp = requests.Response()
        resp.status_code = status_code
        resp_str = resp_body if isinstance(resp_body, str) else json.dumps(resp_body)
        resp.raw = io.BytesIO(resp_str.encode("utf-8"))
        return resp



"""
Base class for the key/value storage abstraction. This abstraction is used to
store keys and values as needed by the various ingestors for state management.
Storage is isolated by namespace so that individual ingestors can operate
independently.
"""

# Standard
from typing import Type
import abc

# Local
from ..factory import FactoryConstructible


class StorageBase(FactoryConstructible):
    __doc__ = __doc__

    # Acceptable value types for storage
    VALUE_TYPE = str | int | float | None

    class StorageNamespaceBase(abc.ABC):
        """A single storage namespace"""

        @abc.abstractmethod
        def set(self, key: str, value: "StorageBase.VALUE_TYPE"):
            """Set the given key/value in this namespace"""

        @abc.abstractmethod
        def get(self, key: str) -> "StorageBase.VALUE_TYPE":
            """Get the value for the given key in this namespace"""

        @abc.abstractmethod
        def pop(self, key: str) -> "StorageBase.VALUE_TYPE":
            """Delete the key from the namespace and return any value that was
            set
            """

    @abc.abstractmethod
    def namespace(str, name: str) -> Type[StorageNamespaceBase]:
        """Get the namespace instance for this name"""



"""
Unit tests for dict-based storage
"""
# Standard
from unittest import mock

# Third Party
import pytest

# Local
from ragnardoc import config
from ragnardoc.storage import storage_factory
from ragnardoc.storage.sqlite_storage import SqliteStorage


def test_factory_construct(scratch_dir):
    """Test that an instance can be constructed from the factory"""
    inst = storage_factory.construct(
        {"type": "sqlite", "config": {"db_path": str(scratch_dir / "storage.db")}}
    )
    assert isinstance(inst, SqliteStorage)


def test_namespace_get_set_pop(scratch_dir):
    """Test that the basic get/set/pop work on a single namespace"""
    inst = storage_factory.construct(
        {"type": "sqlite", "config": {"db_path": str(scratch_dir / "storage.db")}}
    )
    ns = inst.namespace("test")
    assert ns.get("key") is None
    ns.set("key1", 1)
    assert ns.get("key1") == 1
    assert ns.get("key2") is None
    assert ns.pop("key1") == 1
    assert ns.get("key1") is None


def test_multi_namespace(scratch_dir):
    """Test that multiple namespaces can be managed independently"""
    inst = storage_factory.construct(
        {"type": "sqlite", "config": {"db_path": str(scratch_dir / "storage.db")}}
    )
    ns1 = inst.namespace("ns1")
    ns2 = inst.namespace("ns2")
    ns1.set("key", 42)
    assert ns1.get("key") == 42
    assert ns2.get("key") is None


def test_all_types(scratch_dir):
    """Test that the the value types are handled correctly"""
    inst = storage_factory.construct(
        {"type": "sqlite", "config": {"db_path": str(scratch_dir / "storage.db")}}
    )
    ns = inst.namespace("test")
    data = {
        "k1": 1,
        "k2": "two",
        "k3": 3.14,
    }
    for k, v in data.items():
        ns.set(k, v)
    for k, v in data.items():
        assert v == ns.get(k)


def test_invalid_type(scratch_dir):
    """Test that an invalid type is rejected at insertion time"""
    inst = storage_factory.construct(
        {"type": "sqlite", "config": {"db_path": str(scratch_dir / "storage.db")}}
    )
    ns = inst.namespace("test")
    with pytest.raises(TypeError):
        ns.set("key", b"bytes")


def test_ragnardoc_home_db(scratch_dir):
    """Test that non-absolute paths will be prepended by ragnardoc_home"""
    with mock.patch.object(config, "ragnardoc_home", scratch_dir):
        inst = storage_factory.construct(
            {"type": "sqlite", "config": {"db_path": "some_storage.db"}}
        )
        assert inst._db_path == str(scratch_dir / "some_storage.db")


def test_config_defaults(scratch_dir):
    """Test that if db path given, the default is used"""
    with mock.patch.object(config, "ragnardoc_home", scratch_dir):
        inst = storage_factory.construct({"type": "sqlite"})
        assert inst._db_path == str(scratch_dir / "storage.db")



"""
Implementation of the storage abstraction using sqlite3
"""

# Standard
import builtins
import os
import sqlite3

# First Party
import aconfig
import alog

# Local
from .. import config as base_config
from .base import StorageBase

log = alog.use_channel("SQLLSTOR")


class SqliteStorage(StorageBase):
    __doc__ = __doc__

    name = "sqlite"
    config_schema = {
        "type": "object",
        "properties": {
            "db_path": {
                "type": "string",
                "description": "Path or filename for DB file. Non-absolute paths will be placed in RAGNARDOC_HOME",
            },
        },
        "required": ["db_path"],
    }
    config_defaults = {"db_path": "storage.db"}

    def __init__(self, config: aconfig.Config, *_, **__) -> None:
        db_path = config.db_path
        if not os.path.isabs(db_path):
            db_path = os.path.join(base_config.ragnardoc_home, db_path)
        self._db_path = os.path.abspath(db_path)
        log.debug("DB Path: %s", self._db_path)
        os.makedirs(os.path.dirname(self._db_path), exist_ok=True)
        self._conn = sqlite3.connect(self._db_path, check_same_thread=False)

    def __del__(self):
        self._conn.close()

    class StorageSqliteNamespace(StorageBase.StorageNamespaceBase):
        """Implementation of the storage namespace using sqlite3"""

        def __init__(self, name: str, conn: sqlite3.Connection):
            self.name = name
            self._conn = conn

        def set(self, key: str, value: "StorageBase.VALUE_TYPE"):
            type_name = type(value).__name__
            if type_name not in ["str", "int", "float"]:
                raise TypeError(f"Invalid value type: {type_name}")
            with self._conn as conn:
                cursor = conn.cursor()
                self._execute(
                    cursor,
                    f"INSERT OR REPLACE INTO '{self.name}' VALUES (?, ?, ?)",
                    (key, str(value), type_name),
                )
            self._conn.commit()

        def get(self, key: str) -> "StorageBase.VALUE_TYPE":
            with self._conn as conn:
                cursor = conn.cursor()
                self._execute(
                    cursor, f"SELECT * FROM '{self.name}' WHERE key = '{key}' LIMIT 1"
                )
                result = cursor.fetchone()
                if not result:
                    return None
                val = result[1]
                val_type = result[2]
                return self._get_type(val_type)(val)

        def pop(self, key: str) -> "StorageBase.VALUE_TYPE":
            """Delete the key from the namespace and return any value that was
            set
            """
            current_val = self.get(key)
            with self._conn as conn:
                cursor = conn.cursor()
                self._execute(cursor, f"DELETE FROM '{self.name}' WHERE key=?", (key,))
            self._conn.commit()
            return current_val

        @staticmethod
        def _execute(cursor: sqlite3.Cursor, statement: str, args: tuple | None = None):
            log.debug("Executing SQL: %s", statement)
            if args is not None:
                log.debug("Args: %s", args)
                cursor.execute(statement, args)
            else:
                cursor.execute(statement)

        @staticmethod
        def _get_type(type_name: str) -> type:
            return getattr(builtins, type_name)

    def namespace(self, name: str) -> StorageSqliteNamespace:
        ns = self.StorageSqliteNamespace(name, self._conn)
        with self._conn as conn:
            cursor = conn.cursor()
            ns._execute(
                cursor,
                f"""CREATE TABLE IF NOT EXISTS '{name}' (
                key TEXT PRIMARY KEY,
                value TEXT,
                value_type TEXT
            );""",
            )
        return ns



"""
Unit tests for testing the Open WebUI ingestor
"""
# Standard
from contextlib import contextmanager
from unittest import mock
import re
import uuid

# Third Party
import pytest
import requests

# First Party
import aconfig

# Local
from ragnardoc.ingestors.open_webui import OpenWebUIIngestor
from ragnardoc.storage import storage_factory
from ragnardoc.types import Document
from tests.conftest import ServerMockBase

## Helpers #####################################################################


class OpenWebUIMock(ServerMockBase):
    """Mock implementation of OpenWebUI that stores docs in memory"""

    def __init__(self, base_url: str):
        self.base_url = base_url
        self.files = {}
        self.collections = {}

    def _handle_call(self, method: str, url: str, body: dict | None = None):
        """Handle an API call"""
        # NOTE: The trailing slash is there to replicate behavior of Open WebUI
        files_url = f"{self.base_url}/api/v1/files/"
        knowledge_url = f"{self.base_url}/api/v1/knowledge/"

        if method == "post":
            if url == files_url:
                return self._upload(body)
            if m := re.match(f"{files_url}(?P<file_id>[^/]+)/content/update", url):
                return self._update_content(m.group("file_id"), body)
            if url == f"{knowledge_url}create":
                return self._create_collection(body)
            if m := re.match(
                f"{knowledge_url}/(?P<collection_id>[^/]+)/file/update", url
            ):
                collection_id = m.group("collection_id")
                return self._update_collection_file(collection_id, body)
            if m := re.match(f"{knowledge_url}(?P<collection_id>[^/]+)/file/add", url):
                collection_id = m.group("collection_id")
                return self._add_collection_file(collection_id, body)
            if m := re.match(
                f"{knowledge_url}(?P<collection_id>[^/]+)/file/remove", url
            ):
                collection_id = m.group("collection_id")
                return self._remove_collection_file(collection_id, body)

        elif method == "get":
            if m := re.match(f"{files_url}(?P<file_id>[^/]+)", url):
                file_id = m.group("file_id")
                return self._get_file(file_id)
            if url == knowledge_url:
                return self._get_collections()
            if m := re.match(f"{knowledge_url}(?P<collection_id>[^/]+)", url):
                collection_id = m.group("collection_id")
                return self._get_collection(collection_id)

        elif method == "delete":
            if m := re.match(f"{files_url}(?P<file_id>[^/]+)", url):
                file_id = m.group("file_id")
                return self._delete_file(file_id)

        # Any endpoint that doesn't explicitly match redirects back to the web
        # entrypoint, which includes missing trailing slashes
        return self._make_response("<html>this is a web page</html>", 200)

    @classmethod
    def _make_error(cls, status_code: int, msg: str = "") -> requests.Response:
        return cls._make_response({"detail": msg}, status_code)

    ## Handlers ##

    def _upload(self, body: dict) -> requests.Response:
        """Upload always places the doc in custom-documents"""
        try:
            filename, content = body["file"]
        except (IndexError, TypeError):
            return self._make_error(422, "bad file upload")
        uid = str(uuid.uuid4())
        file_value = {"filename": filename, "data": {"content": content}, "id": uid}
        self.files[uid] = file_value
        return self._make_response(file_value)

    def _update_content(self, file_id: str, body: dict) -> requests.Response:
        """Update the content of a file"""
        if file_id not in self.files:
            return self._make_error(404, "file not found")
        try:
            new_content = body["content"]
        except IndexError:
            return self._make_error(422, "bad update")
        self.files[file_id]["data"]["content"] = new_content
        return self._make_response(self.files[file_id])

    def _create_collection(self, body: dict) -> requests.Response:
        """Create a new collection

        NOTE: Open WebUI will happily make a new collection with a duplicate
            name and description!
        """
        uid = str(uuid.uuid4())
        try:
            collection_value = {
                "id": uid,
                "description": body.get("description", ""),
                "name": body["name"],
                "data": {
                    "file_ids": [],
                    "files": [],
                },
            }
            self.collections[uid] = collection_value
            return self._make_response(collection_value)
        except KeyError:
            return self._make_error(422, "bad request")

    def _update_collection_file(
        self, collection_id: str, body: dict
    ) -> requests.Response:
        """Update a file in a collection (reindex)"""
        if collection_id not in self.collections:
            return self._make_error(404, "collection not found")
        try:
            file_id = body["file_id"]
        except KeyError:
            return self._make_error(422, "bad request")
        collection_files = self.collections[collection_id]["files"]
        if file_id not in collection_files:
            return self._make_error(404, "file not in collection")
        return self._make_response(self.collections[collection_id])

    def _add_collection_file(self, collection_id: str, body: dict) -> requests.Response:
        """Add a file to the collection"""
        try:
            file_id = body["file_id"]
        except KeyError:
            return self._make_error(422, "bad request")
        if (collection_value := self.collections.get(collection_id)) is None:
            return self._make_error(404, "collection not found")
        if (file_value := self.files.get(file_id)) is None:
            return self._make_error(404, "file not found")
        file_content = file_value["data"]["content"]
        if file_id in collection_value["data"]["file_ids"] or any(
            col_file["data"]["content"] == file_content
            for col_file in collection_value["data"]["files"]
        ):
            # NOTE: This error is handled specifically
            return self._make_error(
                400,
                "400: Duplicate content detected. Please provide unique content to proceed.",
            )
        collection_value["data"]["file_ids"].append(file_id)
        collection_value["data"]["files"].append(file_value)
        return self._make_response(collection_value)

    def _remove_collection_file(
        self, collection_id: str, body: dict
    ) -> requests.Response:
        """Remove a file from the collection"""
        try:
            file_id = body["file_id"]
        except KeyError:
            return self._make_error(422, "bad request")
        if (collection_value := self.collections.get(collection_id)) is None:
            return self._make_error(404, "collection not found")
        if file_id not in collection_value["data"]["file_ids"]:
            return self._make_error(400, "file not in collection")
        collection_value["data"]["file_ids"].remove(file_id)
        collection_value["data"]["files"] = [
            f for f in collection_value["data"]["files"] if f["id"] != file_id
        ]
        return self._make_response(collection_value)

    def _get_file(self, file_id: str) -> requests.Response:
        """Get a file"""
        if (file_value := self.files.get(file_id)) is None:
            return self._make_error(404, "file not found")
        return self._make_response(file_value)

    def _get_collections(self) -> requests.Response:
        """Get all collections"""
        return self._make_response(self.collections)

    def _get_collection(self, collection_id: str) -> requests.Response:
        """Get a collection"""
        if (collection_value := self.collections.get(collection_id)) is None:
            return self._make_error(404, "collection not found")
        return self._make_response(collection_value)

    def _delete_file(self, file_id: str) -> requests.Response:
        """Delete a file

        NOTE: Open WebUI will happily delete a file that is still in a
            knowledge collection. It seems to remove the entry from "files" but
            not "file_ids"
        """
        if self.files.get(file_id) is None:
            return self._make_error(404, "file not found")
        return self._make_response(self.files.pop(file_id))


@contextmanager
def open_webui_mock_ctx():
    base_url = "http://localhost:5432187"
    mock_server = OpenWebUIMock(base_url)
    with (
        mock.patch("requests.get", mock_server.get),
        mock.patch("requests.post", mock_server.post),
        mock.patch("requests.delete", mock_server.delete),
    ):
        yield mock_server


@pytest.fixture
def open_webui_mock():
    with open_webui_mock_ctx() as mock_server:
        storage = storage_factory.construct({"type": "dict"})
        cfg = aconfig.Config(
            {
                "base_url": mock_server.base_url,
                "apikey": "my-key",
                "knowledge": "ragnardoc_tests",
            },
            override_env_vars=False,
        )
        inst = OpenWebUIIngestor(cfg, "test-inst", storage=storage)
        inst.mock = mock_server
        yield inst


## Tests #######################################################################


def test_anythingllm_init_knowledge(open_webui_mock):
    """Test that the knowledge collection is initialized if needed"""
    assert open_webui_mock._knowledge_id is not None
    assert open_webui_mock._knowledge_id in open_webui_mock.mock.collections


def test_open_webui_ingest_delete(open_webui_mock, mutable_data_dir):
    """Test basic ingestion and deletion"""
    docs = [
        Document.from_file(mutable_data_dir / "sample.txt", mutable_data_dir),
        Document.from_file(
            mutable_data_dir / "sample_docs" / "README.md", mutable_data_dir
        ),
    ]
    open_webui_mock.ingest(docs)

    # Make sure there are two docs uploaded and both were added to the
    # collection
    assert len(open_webui_mock.mock.files) == 2
    assert len(open_webui_mock.mock.collections) == 1
    collection = list(open_webui_mock.mock.collections.values())[0]
    assert all(
        fid in collection["data"]["file_ids"] for fid in open_webui_mock.mock.files
    )

    # Redo ingestion and make sure nothing changes
    open_webui_mock.ingest(docs)
    assert len(open_webui_mock.mock.files) == 2
    assert len(open_webui_mock.mock.collections) == 1
    collection = list(open_webui_mock.mock.collections.values())[0]
    assert all(
        fid in collection["data"]["file_ids"] for fid in open_webui_mock.mock.files
    )

    # Change the content of sample.txt and make sure it gets updated in place
    new_content = "I added some interesting different content!"
    doc_id = [
        doc_id
        for doc_id, doc in open_webui_mock.mock.files.items()
        if doc["filename"].startswith("sample")
    ][0]
    assert open_webui_mock.mock.files[doc_id]["data"]["content"] != new_content
    with open(docs[0].path, "w") as handle:
        handle.write("I added some interesting different content!")
    open_webui_mock.ingest(docs)
    assert open_webui_mock.mock.files[doc_id]["data"]["content"] == new_content

    # Delete the doc and make sure it gets removed from the collection and
    # deleted from files
    open_webui_mock.delete([docs[0]])
    assert len(open_webui_mock.mock.files) == 1
    assert len(open_webui_mock.mock.collections) == 1
    collection = list(open_webui_mock.mock.collections.values())[0]
    assert doc_id not in collection["data"]["file_ids"]
    assert len(collection["data"]["files"]) == 1


def test_open_webui_duplicate_content_ok(open_webui_mock, data_dir):
    """Test that the 400 when re-adding a doc to a collection is handled as ok

    https://github.com/DS4SD/ragnardoc/issues/4
    """
    docs = [Document.from_file(data_dir / "sample.txt", data_dir)]
    open_webui_mock.ingest(docs)

    # Clear the storage to simulate the db being killed
    assert len(open_webui_mock._storage._data) == 1
    open_webui_mock._storage._data.clear()

    # Re-do ingestion and make sure the doc is marked as "done"
    open_webui_mock.ingest(docs)
    assert len(open_webui_mock._storage._data) == 1



"""
Initialize Ragnardoc config for your system
"""
# Standard
import abc
import argparse
import json
import os
import sys

# Third Party
import yaml

# First Party
import alog

# Local
from ... import config
from ...ingestors import ingestor_factory
from ..base import CommandBase
from .anything_llm_initializer import AnythingLLMInitializer
from .open_webui_initializer import OpenWebUIInitializer

log = alog.use_channel("INIT")


class InitCommand(CommandBase):
    __doc__ = __doc__
    name = "init"

    def __init__(self):
        self.auto = False

    def add_args(self, parser: argparse.ArgumentParser):
        """Add all the arguments needed to initialize"""
        parser.add_argument(
            "--home",
            "-H",
            help="Path to where the ragnardoc home should be",
        )
        parser.add_argument(
            "--auto",
            "-a",
            action="store_true",
            help="Run configuration with no interaction",
        )
        parser.add_argument(
            "--overrides",
            "-o",
            help="Path to a file, or a serialized string holding yaml or json config overrides",
        )
        parser.add_argument(
            "--ingestors",
            "-i",
            nargs="*",
            choices=ingestor_factory.registered_types(),
            help="Ingestors to initialize explicitly. If not set, they will be auto-detected.",
        )

    def run(self, args: argparse.Namespace):
        """Run the initialization"""

        # Set the auto bit
        self.auto = args.auto

        # Make sure home exists
        home_dir = args.home or config.ragnardoc_home
        if home_dir != config.ragnardoc_home:
            print(
                "WARNING: Using a custom RAGNARDOC_HOME. Make sure to set RAGNARDOC_HOME in your environment!"
            )
        log.info("Using home: %s", home_dir)
        os.makedirs(home_dir, exist_ok=True)
        user_config_path = os.path.join(home_dir, "config.yaml")

        # Write out the config overrides if given
        user_config = {}
        if args.overrides:
            if os.path.exists(args.overrides):
                with open(args.overrides, "r", encoding="utf-8") as handle:
                    config_override_content = handle.read()
            else:
                config_override_content = args.overrides
            try:
                user_config = json.loads(config_override_content)
            except json.decoder.JSONDecodeError:
                try:
                    with open(args.overrides, "r", encoding="utf-8") as handle:
                        user_config = yaml.safe_load(handle.read())
                except yaml.parser.ParseError as err:
                    log.error("Unable to parse config overrides: %s", err)
                    raise ValueError("Unable to parse config overrides") from err
            if not isinstance(user_config, dict):
                raise ValueError("Config overrides must be a dict")
            if os.path.exists(user_config_path):
                if not self._ask_yes_no("Overwrite current user config", True):
                    sys.exit(1)
                user_config_path_bak = user_config_path + ".bak"
                log.warning(
                    "Backing up previous user config to %s", user_config_path_bak
                )
                os.rename(user_config, user_config_path_bak)

        # Look for ingestors to configure
        ingestors = args.ingestors or []
        if not ingestors:
            print("Detecting available ingestors")
            for ingestor_init in [
                AnythingLLMInitializer,
                OpenWebUIInitializer,
            ]:
                initializer = ingestor_init()
                if initializer.is_installed():
                    if self._ask_yes_no(f"Initialize {initializer.name}", True):
                        ingestor_config = (
                            initializer.initialize_config()
                            if not self.auto
                            else {"type": initializer.name}
                        )
                        user_config.setdefault("ingestion", {}).setdefault(
                            "plugins", []
                        ).append(ingestor_config)

        # Write out all user config if any set up
        if user_config:
            user_config_str = yaml.safe_dump(user_config)
            print("FULL USER CONFIG")
            print()
            print(user_config_str)
            with open(user_config_path, "w", encoding="utf-8") as handle:
                handle.write(user_config_str)

    ## Private Methods ##

    def _ask_yes_no(self, prompt: str, default: bool) -> bool:
        if self.auto:
            return default
        prompt_suffix = "Yn" if default else "yN"
        resp = input(f"{prompt} [{prompt_suffix}]? ")
        if default:
            return resp.lower() != "n"
        return resp.lower() == "y"



"""
Config initializer for Open WebUI
"""
# Third Party
import requests

# Local
from ...ingestors import OpenWebUIIngestor
from ...ingestors.base import Ingestor
from .initializer_base import IngestorInitializerBase


class OpenWebUIInitializer(IngestorInitializerBase):
    __doc__ = __doc__

    def ingestor_class(self) -> type[Ingestor]:
        return OpenWebUIIngestor

    def is_installed(self):
        try:
            resp = requests.get("http://localhost:8080/api/version")
            if resp.status_code == 200:
                # Warn if version is old
                try:
                    # This can fail at json parsing, key indexing, split
                    # unpacking, or int conversion, all of which are caught and
                    # ignored since this is just to do a version check.
                    version = resp.json()["version"]
                    major, minor, patch = (int(elt) for elt in version.split("."))
                    if major != 0 or minor != 5 or patch < 4:
                        print(f"WARNING: Open WebUI version {version} not tested!")
                except (requests.exceptions.JSONDecodeError, KeyError, ValueError):
                    pass
                return True
        except Exception:
            pass
        return False



"""
Unit tests for core types
"""
# Standard
import time

# Local
from ragnardoc import types


def test_document_from_file(txt_data_file, data_dir):
    """Test that a document can be loaded from a file directly"""
    doc = types.Document.from_file(txt_data_file, data_dir, foo=1)
    assert isinstance(doc, types.Document)
    assert doc.path == str(txt_data_file)
    assert doc.converter is None
    assert doc.metadata == {"foo": 1}
    assert doc._content is None


def test_document_from_file_proactive(txt_data_file, data_dir):
    """Test that a document can be loaded from a file directly and proactively
    loaded
    """
    doc = types.Document.from_file(txt_data_file, data_dir, load=True)
    assert doc._content is not None


def test_document_lazy_loading_no_conversion(txt_data_file, data_dir):
    """Test that the document's content is lazily loaded without conversion"""
    doc = types.Document.from_file(txt_data_file, data_dir)
    with open(txt_data_file, "r", encoding="utf-8") as handle:
        expected = handle.read()
    assert doc._content is None
    assert doc.content == expected
    assert doc._content == expected


def test_document_lazy_loading_with_conversion(txt_data_file, data_dir):
    """Test that the document's content is lazily loaded with conversion"""
    expected = "converted!"

    def dummy_converter(ignored):
        return expected

    doc = types.Document.from_file(txt_data_file, data_dir, converter=dummy_converter)
    assert doc._content is None
    assert doc.content == expected
    assert doc._content == expected


def test_set_content():
    """Test that the content property can be set directly"""
    doc = types.Document("some/bad/path", "foo/bar")
    assert doc._content is None
    content = "Hello there!"
    doc.content = content
    assert doc.content == content
    assert doc._content == content


def test_fingerprint_content_change(scratch_dir, data_dir):
    """Test that the doc's fingerprint is computed correctly and mirrors changes
    to the document itself

    NOTE: This test in a previous version did expose a weakness in the current
        fingerprint implementation that uses os.stat: If the content changes to
        something of the exact same length AND the change happens so quickly
        after the initial write that the timestamp precision results in an exact
        equivalent, there's no way to tell the difference with os.stat! This is
        an acceptable limitation given that in ragnardoc, changes would be made
        by users generally who are not operating at this speed.
    """
    doc_path = scratch_dir / "doc.txt"
    content1 = "Hello World"
    # NOTE: This was previously "Hiya world!" which exposed the race condition
    #   with os.stat being exactly equivalent.
    content2 = "Hiya world! How's life these days?"
    with open(doc_path, "w") as handle:
        handle.write(content1)

    # Load from doc with initial content
    doc = types.Document.from_file(doc_path, data_dir)
    fp1 = doc.fingerprint()
    read_content1 = doc.content

    # Make sure the fingerprint doesn't change if the doc hasn't changed
    assert doc.fingerprint() == fp1
    assert doc.content == read_content1 == content1

    # Update the doc content
    with open(doc.path, "w") as handle:
        handle.write(content2)
        handle.flush()

    # Make sure the fingerprint changes and the content is invalidated and
    # re-loaded
    fp2 = doc.fingerprint()
    read_content2 = doc.content
    assert fp2 != fp1
    assert read_content2 != read_content1
    assert doc.content == read_content2 == content2



"""
The core of RAGNARDoc's document crawling and ingestion
"""

# First Party
import aconfig
import alog

# Local
from . import config as default_config
from .ingestors import ingestor_factory
from .scraping import FileScraper
from .storage import storage_factory

log = alog.use_channel("RAGNARDOC")


class RagnardocCore:
    """This is the core class object that maintains the config, scrapers, and
    ingesteors
    """

    def __init__(self, config: aconfig.Config | None = None):
        self.config = config or default_config

        # Construct the storage
        self.storage = storage_factory.construct(self.config.storage)

        # Construct the scraper
        self.scraper = FileScraper(self.storage, self.config.scraping)

        # Construct the ingestors
        self.ingestors = []
        for plugin in self.config.ingestion.plugins:
            try:
                ingestor = ingestor_factory.construct(plugin, storage=self.storage)
                self.ingestors.append(ingestor)
            except Exception as err:
                log.warning(
                    "Failed to construct ingestor %s: %s", plugin.get("type"), err
                )
        log.info(
            "All configured ingestion plugins: %s",
            [entry.name for entry in self.ingestors],
        )

    def ingest(self):
        """Run a single ingestion cycle"""
        log.debug("Initializing scrape")
        with alog.ContextTimer(log.debug, "Done scraping in: "):
            scrape_result = self.scraper.scrape()
        for ingestor in self.ingestors:
            log.debug("Ingesting into %s", ingestor.name)
            if scrape_result.documents:
                with alog.ContextLog(
                    log.info, "Ingesting %d docs", len(scrape_result.documents)
                ):
                    try:
                        ingestor.ingest(scrape_result.documents)
                    except Exception as err:
                        log.warning(
                            "Ingestion failed for ingestor [%s]: err",
                            ingestor.name,
                            err,
                        )
                        log.debug4(err)
            if scrape_result.removed:
                with alog.ContextLog(
                    log.info, "Removing %d docs", len(scrape_result.removed)
                ):
                    try:
                        ingestor.delete(scrape_result.removed)
                    except Exception as err:
                        log.warning(
                            "Deletion failed for ingestor [%s]: err", ingestor.name, err
                        )
                        log.debug4(err)



"""
Common structures for reusable data types
"""

# Standard
from dataclasses import dataclass, field
from pathlib import Path
from typing import Callable
import hashlib
import os

# Type definition of a conversion function that takes the path to a file and
# provides the converted raw text
Converter = Callable[[str], str]


@dataclass
class Document:

    ## Public Attributes ##

    # The path on disk to the document
    path: str
    # The root used to scrape this document
    root: str
    # The title of the document. If unset, the document's basename will be used
    title: str | None = None
    # Additional document metadata
    metadata: dict[str, str | int | float] = field(default_factory=dict)
    # The function that will be used to convert the document to plain text
    converter: Converter | None = None

    ## Private Attributes ##

    # The parsed content of the document. Accessed via content property.
    _content: str | None = None

    # The last computed fingerprint for the document. Used to uniquely identify
    # the content and invalidate the currently read content if needed.
    _last_fingerprint: str | None = None

    ## Properties ##

    @property
    def content(self) -> str:
        """The content of the document will be lazily loaded with conversion
        when first accessed
        """
        self.load()
        return self._content

    @content.setter
    def content(self, value: str):
        self._content = value

    ## Public Methods ##

    def fingerprint(self) -> str | None:
        """The unique fingerprint for this document.

        This is used to determine if/when the document has changed content in
        order to update it where necessary. The current implementation relies on
        a single consistent filesystem as the source of documents, so it uses
        the file metadata rather than a content hash. If the assumption of
        single filesystem changes in the future, this will need to be updated!

        NOTE: Even with the efficient metadata implementation, this is a disk
            operation, so should be used only when needed.

        returns:
            fingerprint (str | None): The unique fingerprint if the file is
            valid, otherwise None (forcing re-read in the future).
        """
        try:
            st = os.stat(self.path)
        except FileNotFoundError:
            return None
        metadata = (
            # File size in bytes
            str(st.st_size),
            # Creation and modification time
            str(st.st_mtime),
            # File permissions
            str(st.st_mode),
        )
        return hashlib.sha256(":".join(metadata).encode()).hexdigest()

    @classmethod
    def from_file(
        cls,
        path: str | Path,
        root: str | Path,
        converter: Converter | None = None,
        load: bool = False,
        **metadata,
    ):
        """Read the Document from the file. By default, it is lazy loaded unless
        load is True.
        """
        inst = cls(
            path=str(path), root=str(root), converter=converter, metadata=metadata
        )
        if load:
            inst.load()
        return inst

    def load(self):
        """If content is not yet set or is invalid, load and convert it"""
        fingerprint = self.fingerprint()
        if self._content is None or fingerprint != self._last_fingerprint:
            # NOTE: This is _not_ thread safe! If ingestion supports threading
            #   (even with the GIL), this will need a lock
            self._last_fingerprint = fingerprint
            if self.converter:
                self._content = self.converter(self.path)
            else:
                with open(self.path, encoding="utf-8") as handle:
                    self._content = handle.read()


@dataclass
class ScrapeResult:
    """The result of a single scrape is a set of documents that exist and a set
    that have been removed
    """

    documents: list[Document]
    removed: list[Document]



"""
Ingestor for AnythingLLM
https://anythingllm.com/

Currently, the AnythingLLM API is fairly hidden and does not support update
operations on documents, so instead we add a timestamp to the docs when uploaded
to indicate to the user their update time.
"""

# Standard
from datetime import datetime
import os

# Third Party
import requests

# First Party
import aconfig
import alog

# Local
from ..storage import StorageBase
from ..types import Document
from .base import Ingestor

log = alog.use_channel("ANYTHINGLLM")


class AnythingLLMIngestor(Ingestor):
    __doc__ = __doc__

    name = "anything-llm"
    config_schema = {
        "type": "object",
        "properties": {
            "base_url": {
                "type": "string",
                "description": "The base URL for the AnythingLLM server",
            },
            "apikey": {
                "type": "string",
                "description": "The developer API Key for the AnythingLLM server",
            },
            "root_folder": {
                "type": "string",
                "description": "The root folder name in the AnythingLLM doc tree",
            },
            "workspaces": {
                "type": "array",
                "items": {
                    "type": "string",
                },
                "description": "List of workspaces to add documents to",
            },
        },
        "required": ["apikey"],
    }
    config_defaults = {
        "base_url": "http://localhost:3001",
        "root_folder": "ragnardoc",
        "workspaces": [],
    }

    def __init__(
        self,
        config: aconfig.Config,
        instance_name: str,
        *,
        storage: StorageBase,
    ):
        # All the API URLs we'll use for easy reference
        self._base_url = config.base_url
        api_base = f"{self._base_url}/api/v1"
        self._upload_url = f"{api_base}/document/raw-text"
        self._move_url = f"{api_base}/document/move-files"
        self._create_folder_url = f"{api_base}/document/create-folder"
        self._workspaces_url = f"{api_base}/workspaces"
        self._workspace_details_url = f"{api_base}/workspace"
        self._doc_delete_url = f"{api_base}/system/remove-documents"

        self._root_folder = config.root_folder

        self._headers = {
            "Authorization": f"Bearer {config.apikey}",
        }

        # Scoped storage for re-ingestion fingerprint cache
        self._storage = storage.namespace(self.name + instance_name)

        # Get the workspace slugs for all configured workspaces
        self._workspace_slugs = {
            ws["name"]: ws["slug"]
            for ws in self._get_workspaces()
            if ws["name"] in config.workspaces
        }
        log.debug("Workspace slugs: %s", self._workspace_slugs)
        missing_workspaces = [
            ws for ws in config.workspaces if ws not in self._workspace_slugs
        ]
        if missing_workspaces:
            log.warning("Workspace(s) not found: %s", missing_workspaces)

    #######################
    ## Interface Methods ##
    #######################

    def ingest(self, documents: list[Document]):
        """Ingest the documents, updating existing docs as necessary"""
        # Ensure the base ragnardoc folder exists
        self._ensure_directory_path(self._root_folder)

        uploaded_docs = []
        for doc in documents:

            # Check to see if this doc has changed since last ingesting
            fingerprint = doc.fingerprint()
            if self._storage.get(doc.path) == fingerprint:
                log.debug("Document [%s] has not changed. Not uploading", doc.path)
                continue

            # Ensure the latest content is current
            try:
                doc_content = doc.content
            except Exception as err:
                log.debug("Unable to parse document %s: %s", doc.path, err)
                log.debug4(err, exc_info=True)
                continue

            # Do the raw ingestion into custom-documents
            title = self._get_doc_title(doc)
            log.info("Ingesting document: %s", title)
            resp = requests.post(
                self._upload_url,
                headers=self._headers,
                json={
                    "textContent": doc_content,
                    "metadata": {
                        "title": title,
                    },
                },
            )
            log.debug("Upload response code: %d", resp.status_code)
            log.debug2(resp.text)

            # Parse the response json if possible
            try:
                resp_json = resp.json()
            except requests.exceptions.JSONDecodeError:
                resp_json = {}

            # Handle errors
            if resp.status_code != 200:
                log.warning(
                    "Failed to upload document %s: %s",
                    doc.path,
                    resp_json.get("message"),
                )
                log.debug(resp.text)
                continue

            # Move the document to the root folder. Here, we use a name that is
            # unique to the doc, but _not_ unique to the upload. This
            # approximates "update" semantics.
            try:
                upload_location = resp_json["documents"][0]["location"]
            except KeyError:
                log.warning("No location found in first document!")
                continue
            target_location = self._get_doc_location(doc)
            move_resp = requests.post(
                self._move_url,
                headers=self._headers,
                json={"files": [{"from": upload_location, "to": target_location}]},
            )
            if move_resp.status_code != 200:
                log.warning("Failed to move document %s to correct location", doc.path)
                continue

            # If everything successful so far, store the fingerprint for future
            # re-ingest checks
            self._storage.set(doc.path, fingerprint)
            uploaded_docs.append(target_location)

        # Update the workspaces with the uploaded docs
        for workspace_name, workspace_slug in self._workspace_slugs.items():
            log.debug2("Updating docs in workspace %s", workspace_name)
            self._update_docs_in_workspace(uploaded_docs, workspace_slug)

    def delete(self, documents: list[Document]):
        """Currently, there is no good way to delete docs!"""
        # Remove all documents from workspaces where they're indexed
        doc_locations = [self._get_doc_location(doc) for doc in documents]
        for workspace_name, workspace_slug in self._workspace_slugs.items():
            workspace_docs = self._workspace_doc_paths(workspace_slug)
            if docs_to_remove := [
                doc_loc for doc_loc in doc_locations if doc_loc in workspace_docs
            ]:
                log.debug(
                    "Removing docs from workspace %s: %s",
                    workspace_name,
                    docs_to_remove,
                )
                remove_resp = requests.post(
                    f"{self._workspace_details_url}/{workspace_slug}/update-embeddings",
                    headers=self._headers,
                    json={"deletes": docs_to_remove},
                )
                remove_resp.raise_for_status()

        # Fully delete the docs
        delete_resp = requests.delete(
            self._doc_delete_url,
            headers=self._headers,
            json={"names": doc_locations},
        )
        delete_resp.raise_for_status()

        # Clear out the fingerprints
        for doc in documents:
            self._storage.pop(doc.path)

    #####################
    ## Private Methods ##
    #####################

    @staticmethod
    def _get_doc_title(doc: Document) -> str:
        """Get the title to be used for this doc in Anything LLM's storage.

        NOTE: Anything LLM does not support nested directories so we need to
            flatten the path with '--' characters
        """
        raw_title = os.path.relpath(
            doc.path, os.path.dirname(os.path.abspath(doc.root))
        )
        title = raw_title.replace(os.sep, "--")
        log.debug2("Title for %s: %s", doc.path, title)
        return title

    def _get_doc_location(self, doc: Document) -> str:
        """Get the location where the doc will be stored in AnythingLLM's doc
        cache dir
        """
        return os.path.join(self._root_folder, self._get_doc_title(doc) + ".json")

    def _ensure_directory_path(self, dirpath: str):
        """Create the given document directory path exists"""
        resp = requests.post(
            self._create_folder_url,
            headers=self._headers,
            json={"name": dirpath},
        )
        # If it's a 500 because the dir already exists, that's ok! If any of the
        # logic to check this fails and raises, that's not ok and it _should_
        # raise.
        if resp.status_code == 500:
            try:
                resp_body = resp.json()
                if (
                    not resp_body["success"]
                    and "already exists" in resp_body["message"]
                ):
                    log.debug4("Directory %s already exists", dirpath)
                    return
            except Exception as err:
                log.error("Failed to set up directory %s: %s", dirpath, err)
                raise
        resp.raise_for_status()

    def _get_workspaces(self) -> list[dict]:
        """Get info about all workspaces"""
        resp = requests.get(self._workspaces_url, headers=self._headers)
        resp.raise_for_status()
        return resp.json()["workspaces"]

    def _workspace_doc_paths(self, workspace_slug: str) -> list[str]:
        """Get the list of doc paths in a given workspace"""
        resp = requests.get(
            f"{self._workspace_details_url}/{workspace_slug}",
            headers=self._headers,
        )
        resp.raise_for_status()
        return [doc["docpath"] for doc in resp.json()["workspace"][0]["documents"]]

    def _update_docs_in_workspace(self, uploaded_docs: list[str], workspace_slug: str):
        """Attempt to update the given workspace with the given uploaded docs"""
        try:
            existing_docs = self._workspace_doc_paths(workspace_slug)

            # Remove any existing docs that have been updated
            if updated_docs := [doc for doc in uploaded_docs if doc in existing_docs]:
                log.debug(
                    "Removing updated docs from %s: %s", workspace_slug, updated_docs
                )
                remove_resp = requests.post(
                    f"{self._workspace_details_url}/{workspace_slug}/update-embeddings",
                    headers=self._headers,
                    json={"deletes": updated_docs},
                )
                remove_resp.raise_for_status()

            # Add all the docs as new to trigger indexing
            add_resp = requests.post(
                f"{self._workspace_details_url}/{workspace_slug}/update-embeddings",
                headers=self._headers,
                json={"adds": uploaded_docs},
            )
            add_resp.raise_for_status()
        except Exception as err:
            log.warning("Failed to update docs for %s: %s", workspace_slug, err)
            log.debug(err, exc_info=True)



"""
Base class for initializers that can be used to auto-initialize ingestion plugin
connections
"""

# Standard
import abc

# Third Party
import yaml

# Local
from ...ingestors.base import Ingestor


class IngestorInitializerBase(abc.ABC):
    __doc__ = __doc__

    @property
    def name(self) -> str:
        return self.ingestor_class().name

    @abc.abstractmethod
    def ingestor_class(self) -> type[Ingestor]:
        """Get the ingestor class for this initializer"""

    @abc.abstractmethod
    def is_installed(self):
        """Determine whether the ingestor is installed"""

    def initialize_config(self) -> dict:
        """Initialize the config for this ingestor"""
        ingestor_class = self.ingestor_class()
        return_dict = {"type": ingestor_class.name}
        config_dict = return_dict.setdefault("config", {})

        # Print out the config schema
        print("CONFIG SCHEMA")
        print(yaml.safe_dump(ingestor_class.config_schema))
        print()
        print("CONFIG_DEFAULTS")
        print(yaml.safe_dump(ingestor_class.config_defaults))
        print()
        # TODO: Figure out required params with no defaults and always ask for
        # them
        while True:
            if input("Enter a config value [Yn]? ").lower() == "n":
                break
            print("HINT: Join nested keys with '.' characters.")
            print("HINT: Lists will always be appended to.")
            key = input("Key: ").strip()
            if not key:
                continue
            val = input("Value: ").strip()
            try:
                self._recursive_set(config_dict, ingestor_class.config_schema, key, val)
            except KeyError as err:
                print(f"Could not set [{key} = {val}]: {err}")
        return return_dict

    @classmethod
    def _recursive_set(cls, config_dict: dict, schema: dict, key: str, val: str):
        """Recursive setter function to add the given key inside the given dict"""
        top_key, _, key_remainder = key.partition(".")
        prop = schema.get("properties", {}).get(top_key, {})
        val_type = prop.get("type")
        if top_key == key:
            if val_type is None:
                raise KeyError(f"Key not in schema: {key}")
            converted = cls._convert_schema_type(val, prop)
            if isinstance(converted, list):
                config_dict.setdefault(key, []).extend(converted)
            else:
                config_dict[key] = converted
        else:
            cls._recursive_set(
                config_dict.setdefault(top_key, {}), prop, key_remainder, val
            )

    @classmethod
    def _convert_schema_type(cls, val: str, prop: dict) -> str | int | float | list:
        val_type = prop.get("type")
        if val_type == "number":
            try:
                return int(val)
            except ValueError:
                return float(val)
        if val_type == "string":
            return val
        if val_type == "array":
            items = prop.get("items", {})
            item_type = items.get("type")
            if item_type in ["object", "array"]:
                raise KeyError(
                    "Cannot set nested objects or arrays here. Use manual config overrides!"
                )
            return [cls._convert_schema_type(val, items)]
        raise ValueError(f"Unsupported type: {val_type}")



"""
Unit tests for the start command
"""
# Standard
from datetime import timedelta
from unittest import mock
import argparse
import threading
import time

# Third Party
import pytest

# First Party
import aconfig

# Local
from ragnardoc.cli.start import StartCommand


@pytest.mark.parametrize(
    ["time_str", "expected_delta"],
    [
        ("35s", timedelta(seconds=35)),
        ("1d  2h  35s", timedelta(seconds=35 + 2 * 60 * 60 + 60 * 60 * 24)),
        ("16s 6h", timedelta(seconds=16 + 6 * 60 * 60)),
        ("2hours 1minute", timedelta(seconds=60 + 2 * 60 * 60)),
        ("0.5s", timedelta(seconds=0.5)),
    ],
)
def test_parse_time(time_str, expected_delta):
    """Test that time parsing works for various combinations"""
    assert StartCommand._parse_time(time_str) == expected_delta


@pytest.mark.parametrize("time_str", ["", "  ", "1 d", "1w"])
def test_parse_time_invalid(time_str):
    """Test that ValueError is raised for invalid time strings"""
    with pytest.raises(ValueError):
        StartCommand._parse_time(time_str)


@mock.patch("subprocess.run")
def test_run(run_mock):
    """Test that running the command launches the infinite loop correctly"""
    cmd = StartCommand()
    args = aconfig.Config({"period": "0.1s"}, override_env_vars=False)
    run_thread = threading.Thread(target=cmd.run, args=(args,))
    run_thread.start()
    time.sleep(0.05)
    cmd.stop()
    run_thread.join()
    run_mock.assert_called_once()


def test_add_args():
    """Test that the command adds the expected arguments"""
    parser = argparse.ArgumentParser()
    StartCommand().add_args(parser)
    args = parser.parse_args([])
    assert hasattr(args, "period")



"""
Central configuration for the tool
"""
# Standard
import os

# First Party
import aconfig

# Local
from .merge import merge_configs

BASE_CONFIG_PATH = os.path.join(os.path.dirname(__file__), "config.yaml")


def _initialize_config() -> aconfig.ImmutableConfig:
    # Parse the base config with env var overrides
    config = aconfig.Config.from_yaml(BASE_CONFIG_PATH, override_env_vars=True)

    # Make sure '~' is expanded correctly
    config.ragnardoc_home = os.path.expanduser(config.ragnardoc_home)

    # Merge in user overrides
    user_config = os.path.join(config.ragnardoc_home, "config.yaml")
    if os.path.exists(user_config):
        config = merge_configs(config, aconfig.Config.from_yaml(user_config))

    # Return the immutable view
    return aconfig.ImmutableConfig(config)


# Global config object
config_instance = _initialize_config()

# Attribute access
def __getattr__(key: str) -> any:
    return getattr(config_instance, key)



"""
Central module for all CLI commands
"""

# Local
from .add import AddCommand
from .common import add_common, use_common
from .init import InitCommand
from .run import RunCommand
from .start import StartCommand

all_commands = {
    cmd.name: cmd
    for cmd in [
        AddCommand,
        InitCommand,
        RunCommand,
        StartCommand,
    ]
}



"""
Ingestor for Open WebUI
https://openwebui.com/
"""

# Standard
import json
import os

# Third Party
import requests

# First Party
import aconfig
import alog

# Local
from ..storage import StorageBase
from ..types import Document
from .base import Ingestor

log = alog.use_channel("OPENWEBUI")


class OpenWebUIIngestor(Ingestor):
    __doc__ = __doc__

    name = "open-webui"
    config_schema = {
        "type": "object",
        "properties": {
            "base_url": {
                "type": "string",
                "description": "The base URL for the AnythingLLM server",
            },
            "apikey": {
                "type": "string",
                "description": "The developer API Key for the AnythingLLM server",
            },
            "knowledge": {
                "type": "string",
                "description": "The knowledge collection to place the docs in",
            },
        },
        "required": ["apikey"],
    }
    config_defaults = {
        "base_url": "http://localhost:8080",
        "knowledge": "ragnardoc",
    }

    def __init__(
        self,
        config: aconfig.Config,
        instance_name: str,
        *,
        storage: StorageBase,
    ):
        # All the API URLs we'll use for easy reference
        # NOTE: Open WebUI is very sensitive to the trailing slash! Some
        #   endpoints require it, others require it to not be present.
        self._base_url = config.base_url
        self._files_url = f"{self._base_url}/api/v1/files/"
        self._knowledge_url = f"{self._base_url}/api/v1/knowledge/"

        # Common headers for all requests
        self._headers = {
            "Authorization": f"Bearer {config.apikey}",
        }

        # Central knowledge collection for ragnardoc
        self._knowledge_id = self._ensure_knowledge_collection(config.knowledge)
        self._knowledge_collection_url = f"{self._knowledge_url}{self._knowledge_id}"

        # Scoped storage for re-ingestion fingerprint cache
        self._storage = storage.namespace(self.name + instance_name)

    #######################
    ## Interface Methods ##
    #######################

    def ingest(self, documents: list[Document]):
        """Ingest the documents, updating existing docs as necessary"""
        uploaded_doc_ids = []
        for doc in documents:
            log.debug3("Checking %s for Open WebUI ingestion", doc.path)

            # Check to see if this doc has changed since last ingesting
            fingerprint = doc.fingerprint()
            stored_fingerprint, file_id = None, None
            if stored_content := self._get_file_storage(doc.path):
                stored_fingerprint = stored_content["fingerprint"]
                file_id = stored_content["id"]
            if stored_fingerprint == fingerprint:
                log.debug("Document [%s] has not changed. Not uploading", doc.path)
                continue

            # Ensure the latest content is current
            try:
                doc_content = doc.content
            except Exception as err:
                log.debug("Unable to parse document %s: %s", doc.path, err)
                log.debug4(err, exc_info=True)
                continue

            # Check to see if the file already exists in Open WebUI
            file_exists = False
            if file_id:
                resp = requests.get(f"{self._files_url}{file_id}")
                file_exists = resp.status_code == 200

            # If the file already exists, update its content
            if file_exists:
                log.debug2("Updating existing file %s with id %s", doc.path, file_id)
                resp = requests.post(
                    f"{self._files_url}{file_id}/content/update",
                    headers=self._headers,
                    json={"content": doc_content},
                )
                if resp.status_code != 200:
                    log.warning(
                        "Failed to update doc %s: [%d] %s",
                        doc.path,
                        resp.status_code,
                        resp.text,
                    )
                    continue

                # Update this doc to the knowledge collection
                resp = requests.post(
                    f"{self._knowledge_collection_url}/file/update",
                    headers=self._headers,
                    json={"file_id": file_id},
                )
                if resp.status_code != 200:
                    log.warning(
                        "Failed to update document %s with id %s in knowledge collection: %s",
                        doc.path,
                        file_id,
                        resp.text,
                    )
                    continue

            # Otherwise, upload it
            else:
                log.debug2("Uploading new document: %s", doc.path)
                # Get the filename that will be used in Open WebUI
                filename = self._get_filename(doc)
                resp = requests.post(
                    f"{self._files_url}",
                    headers=self._headers,
                    files={"file": (filename, doc_content)},
                )
                if resp.status_code != 200:
                    log.warning("Failed to upload doc %s: %s", doc.path, resp.text)
                try:
                    resp_body = resp.json()
                    file_id = resp_body["id"]
                except (requests.exceptions.JSONDecodeError, KeyError) as err:
                    log.warning("Failed to get file_id from upload response: %s", err)
                    log.debug4(err, exc_info=True)
                    continue

                # Add this doc to the knowledge collection
                log.debug2(
                    "Adding doc with id %s to knowledge collection %s",
                    file_id,
                    self._knowledge_id,
                )
                resp = requests.post(
                    f"{self._knowledge_collection_url}/file/add",
                    headers=self._headers,
                    json={"file_id": file_id},
                )
                if (
                    resp.status_code == 400
                    and "Duplicate content detected" in resp.text
                ):
                    log.debug("Doc %s already added to knowledge collection", doc.path)
                elif resp.status_code != 200:
                    log.warning(
                        "Failed to add document %s with id %s to knowledge collection: %s",
                        doc.path,
                        file_id,
                        resp.text,
                    )
                    continue

            # Mark this doc as successfully uploaded
            uploaded_doc_ids.append(file_id)

            # Update the storage to reflect this file
            self._set_file_storage(doc.path, fingerprint, file_id)

    def delete(self, documents: list[Document]):
        """Currently, there is no good way to delete docs!"""
        # Get the file_ids from storage for each doc
        doc_infos = [self._get_file_storage(doc.path) for doc in documents]
        doc_file_ids = {
            stored_content["id"] for stored_content in doc_infos if stored_content
        }
        for file_id in doc_file_ids:

            # Remove from knowledge collection
            resp = requests.post(
                f"{self._knowledge_collection_url}/file/remove",
                headers=self._headers,
                json={"file_id": file_id},
            )
            if resp.status_code != 200:
                log.warning(
                    "Failed to remove doc with id %s from knowledge collection", file_id
                )
                continue

            # Delete it
            resp = requests.delete(f"{self._files_url}{file_id}", headers=self._headers)
            if resp.status_code != 200:
                log.warning("Failed to delete doc with id %s", file_id)

    #####################
    ## Private Methods ##
    #####################

    def _get_file_storage(self, path: str) -> dict | None:
        if stored_content := self._storage.get(path):
            return json.loads(stored_content)

    def _set_file_storage(self, path: str, fingerprint: str, file_id: str):
        self._storage.set(
            path,
            json.dumps(
                {
                    "fingerprint": fingerprint,
                    "id": file_id,
                }
            ),
        )

    def _ensure_knowledge_collection(self, knowledge_collection: str) -> str:
        """Create the given knowledge collection if needed and return the id"""
        # Get all knowledge collections and look for one matching this name
        resp = requests.get(self._knowledge_url, headers=self._headers)
        resp.raise_for_status()
        all_knowledge_collections = {col["name"]: col["id"] for col in resp.json()}
        if knowledge_id := all_knowledge_collections.get(knowledge_collection):
            log.debug2(
                "Knowledge collection [%s] already exists with id: %s",
                knowledge_collection,
                knowledge_id,
            )
            return knowledge_id

        # If not found above, create it and return the ID
        resp = requests.post(
            f"{self._knowledge_url}create",
            headers=self._headers,
            json={
                "name": knowledge_collection,
                "description": "Documents ingested with RAGNARDoc",
            },
        )
        resp.raise_for_status()
        return resp.json()["id"]

    @staticmethod
    def _get_filename(doc: Document) -> str:
        """The file name will be formatted with the actual file name at the
        beginning, followed by a qualifying path relative to the root, followed
        by a file extension. This is to allow the user to easily reference the
        file by name, but still disambiguate by path and to avoid confusing Open
        WebUI with non-text file types.
        """
        abs_path = os.path.abspath(doc.path)
        file_name = os.path.basename(abs_path)
        if doc.converter:
            ext = ".md"
        else:
            file_name, ext = os.path.splitext(file_name)
        rel_path = os.path.relpath(
            os.path.dirname(abs_path), os.path.dirname(os.path.abspath(doc.root))
        ).replace(os.sep, "--")
        filename = f"{file_name} ({rel_path}){ext}"
        return filename



#!/usr/bin/env python
"""
RAGNARDoc is a tool that runs natively on a developer workstation to
automatically ingest local documents into a variety of RAG applications. It can
operate as a CLI for direct ingestion or as a service for background ingestion
and synchronization (Coming soon!)
"""

# Standard
import argparse

# First Party
import alog

# Local
from . import cli, config
from ._version import __version__

log = alog.use_channel("MAIN")


def main():
    # Initial parser to parse the command
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("command", nargs="?", choices=sorted(cli.all_commands.keys()))
    parser.add_argument(
        "--version", "-v", action="store_true", help="Show the library version"
    )
    initial_args = parser.parse_known_args()[0]

    # Handle version
    if initial_args.version:
        print(__version__)
        return 0

    # Full parser to parse the command arguments
    command = initial_args.command
    cmd_cls = cli.all_commands.get(command)
    parser = argparse.ArgumentParser(description=parser.description)
    parser.add_argument("command", choices=sorted(cli.all_commands.keys()))
    cli.add_common(parser)

    # Add the args for the specific command
    if cmd_cls:
        cmd_inst = cmd_cls()
        cmd_inst.add_args(
            parser.add_argument_group(cmd_cls.name, description=cmd_cls.__doc__)
        )

    # Parse all args and handle common setup
    args = parser.parse_args()
    cli.use_common(args)

    # Run the command
    log.info("RAGNARDoc is running command %s", command)
    log.debug4("Full config: %s", config.config_instance)
    cmd_inst.run(args)


if __name__ == "__main__":
    main()



"""
Extensible set of storage implementations
"""

# Local
from ..factory import ImportableFactory
from .base import StorageBase
from .dict_storage import DictStorage
from .sqlite_storage import SqliteStorage

storage_factory = ImportableFactory("storage")
storage_factory.register(DictStorage)
storage_factory.register(SqliteStorage)



"""
Common CLI argument setup
"""

# Standard
import argparse

# First Party
import alog

# Local
from .. import config


def add_common(parser: argparse.ArgumentParser):
    """Add common arguments"""
    # Logging
    log_section = parser.add_argument_group(
        "log", description="Output logging configuration"
    )
    log_section.add_argument(
        "--log-level", "-l", default=config.log_level, help="Default log channel level"
    )
    log_section.add_argument(
        "--log-filters",
        "-lf",
        default=config.log_filters,
        help="Per-channel log filters",
    )
    log_section.add_argument(
        "--log-json",
        "-lj",
        action="store_true",
        default=config.log_json,
        help="Log output as JSON",
    )
    log_section.add_argument(
        "--log-thread-id",
        "-lt",
        action="store_true",
        default=config.log_thread_id,
        help="Include the thread ID in log header",
    )


def use_common(args: argparse.Namespace):
    """Use the common arguments"""
    # Configure logging
    alog.configure(
        default_level=args.log_level,
        filters=args.log_filters,
        formatter="json" if args.log_json else "pretty",
        thread_id=args.log_thread_id,
    )



"""
Module for scraping files to ingest
"""
# Standard
import json
import os
import re

# First Party
import aconfig
import alog

# Local
from .storage import StorageBase
from .types import Document, ScrapeResult

log = alog.use_channel("SCRAPING")


class FileScraper:

    _scrape_cache_key = "scrape_cache"

    def __init__(self, storage: StorageBase, config: aconfig.Config):

        # NOTE: Local import to avoid slow imports for non-run commands
        # Third Party
        from docling.document_converter import DocumentConverter

        # Load the docling converter
        with alog.ContextTimer(log.debug, "Loaded doc converter in: "):
            self.converter = DocumentConverter()

        # Figure out the paths to scrape from
        self.roots = [os.path.expanduser(root) for root in config.roots]

        # Save the configured set of raw text types that don't need conversion
        self.raw_text_extensions = config.raw_text_extensions

        # Store the include/exclude patterns
        self.include_paths = config.include.paths
        self.include_regexprs = [
            re.compile(expr) for expr in config.include.regexprs or [".*"]
        ]
        self.exclude_paths = config.exclude.paths
        self.exclude_regexprs = [re.compile(expr) for expr in config.exclude.regexprs]

        # Scoped storage for detecting deletions
        self._storage = storage.namespace("__core_scraping__")
        self._auto_delete = config.auto_delete

    def scrape(self) -> ScrapeResult:
        """Scrape the given path"""
        files_to_ingest = {}
        for root in self.roots:
            log.debug("Scraping root: %s", root)
            for parent, _, files in os.walk(root):
                log.debug2("Scraping contents of %s", parent)
                for fname in files:
                    full_path = os.path.join(parent, fname)
                    if (
                        self._match_paths(full_path, self.include_paths)
                        or self._match_regexprs(full_path, self.include_regexprs)
                    ) and not (
                        self._match_paths(full_path, self.exclude_paths)
                        or self._match_regexprs(full_path, self.exclude_regexprs)
                    ):
                        files_to_ingest.setdefault(root, []).append(full_path)
        all_ingest_paths = [doc for root in files_to_ingest.values() for doc in root]
        for include_path in self.include_paths:
            if include_path not in all_ingest_paths:
                files_to_ingest.setdefault(os.path.sep, []).append(include_path)
        log.debug4("All docs to ingest: %s", files_to_ingest)

        # Construct the docs (with lazy loading)
        output_docs = {}
        for root, root_files in files_to_ingest.items():
            for fname in root_files:
                is_raw_text = self._is_raw_text_type(fname)
                log.debug2(
                    "Doc %s %s raw text", fname, "IS" if is_raw_text else "IS NOT"
                )
                converter = None if is_raw_text else self._convert_doc
                # Sometimes docs are found multiple times with redundant roots
                # or includes
                if fname not in output_docs:
                    output_docs[fname] = Document.from_file(
                        path=fname, root=root, converter=converter
                    )

        # Detect deleted docs
        this_scrape_data = {doc.path: doc.root for doc in output_docs.values()}
        deleted_docs = []
        if self._auto_delete and (
            last_scrape_data := self._storage.get(self._scrape_cache_key)
        ):
            last_scrape = json.loads(last_scrape_data)
            deleted_docs = [
                Document(path=doc_path, root=doc_root)
                for doc_path, doc_root in last_scrape.items()
                if doc_path not in this_scrape_data
            ]

        # Add this scrape to the last scrape cache
        self._storage.set(self._scrape_cache_key, json.dumps(this_scrape_data))

        # Return the full result of the scrape
        return ScrapeResult(documents=list(output_docs.values()), removed=deleted_docs)

    ## Impl ##

    @staticmethod
    def _match_paths(candidate: str, paths: list[str]) -> bool:
        return any(path == candidate for path in paths)

    @staticmethod
    def _match_regexprs(candidate: str, exprs: list[re.Pattern]) -> bool:
        return any(expr.match(candidate) for expr in exprs)

    def _is_raw_text_type(self, candidate: str) -> bool:
        return (
            os.path.splitext(candidate)[1].lower().lstrip(".")
            in self.raw_text_extensions
        )

    def _convert_doc(self, fname: str) -> Document | None:
        converted = self.converter.convert(fname)
        return converted.document.export_to_markdown()



"""
Add a document or directory to RAGNARDoc
"""
# Standard
import argparse
import os

# Third Party
import yaml

# First Party
import alog

# Local
from .. import config
from .base import CommandBase

log = alog.use_channel("ADD")


class AddCommand(CommandBase):
    __doc__ = __doc__
    name = "add"

    def add_args(self, parser: argparse.ArgumentParser):
        """Add arbitrary documents"""
        parser.add_argument(
            "paths",
            nargs="+",
            help="Paths to files or directories to add to RAGNARDoc",
        )

    def run(self, args: argparse.Namespace):
        """Add the docs to the config"""
        # Read the existing user config
        user_config_path = os.path.join(config.ragnardoc_home, "config.yaml")
        user_config = {}
        if os.path.exists(user_config_path):
            with open(user_config_path, "r", encoding="utf-8") as handle:
                user_config = yaml.safe_load(handle)

        # Add each path to either the roots or to the includes
        scraping = user_config.setdefault("scraping", {})
        for path in args.paths:
            path = self._normalize_path(path)
            if os.path.isdir(path):
                if path not in (current_roots := scraping.setdefault("roots", [])):
                    current_roots.append(path)
            elif os.path.exists(path):
                if path not in (
                    current_paths := scraping.setdefault("include", {}).setdefault(
                        "paths", []
                    )
                ):
                    current_paths.append(path)
            else:
                print(f"WARNING: Ignoring non-existing path [{path}]")

        # Write the config back out
        user_config_str = yaml.safe_dump(user_config)
        log.debug("FULL CONFIG\n%s", user_config_str)
        with open(user_config_path, "w", encoding="utf-8") as handle:
            handle.write(user_config_str)

    def _normalize_path(self, path: str) -> str:
        return os.path.normpath(os.path.realpath(os.path.expanduser(path)))



"""
The run command runs a single ragnardoc ingestion based on the available config
"""
# Standard
import argparse

# First Party
import alog

# Local
from .. import config
from ..core import RagnardocCore
from .base import CommandBase

log = alog.use_channel("RUN")


class RunCommand(CommandBase):
    __doc__ = __doc__
    name = "run"

    def add_args(self, *_, **__):
        """There are no additional args to the run command"""

    def run(self, args: argparse.Namespace):
        """Perform the single run"""
        instance = RagnardocCore(config)
        with alog.ContextTimer(log.info, "Finished ingestion in: "):
            instance.ingest()



"""
Unit tests for dict-based storage
"""

# Local
from ragnardoc.storage import storage_factory
from ragnardoc.storage.dict_storage import DictStorage


def test_factory_construct():
    """Test that an instance can be constructed from the factory"""
    inst = storage_factory.construct({"type": "dict"})
    assert isinstance(inst, DictStorage)


def test_namespace_get_set_pop():
    """Test that the basic get/set/pop work on a single namespace"""
    inst = DictStorage()
    ns = inst.namespace("test")
    assert ns.get("key") is None
    ns.set("key1", 1)
    assert ns.get("key1") == 1
    assert ns.get("key2") is None
    assert ns.pop("key1") == 1
    assert ns.get("key1") is None


def test_multi_namespace():
    """Test that multiple namespaces can be managed independently"""
    inst = DictStorage()
    ns1 = inst.namespace("ns1")
    ns2 = inst.namespace("ns2")
    ns1.set("key", 42)
    assert ns1.get("key") == 42
    assert ns2.get("key") is None



"""
Implementation of the storage abstraction using an in-memory dict
"""

# Local
from .base import StorageBase


class DictStorage(StorageBase):
    __doc__ = __doc__

    name = "dict"
    config_schema = {"type": "object"}

    def __init__(self, *_, **__):
        self._data = {}

    class DictStorageNamespace(StorageBase.StorageNamespaceBase):
        """Single dict namespace"""

        def __init__(self, name: str, parent: "DictStorage"):
            self._data = parent._data.setdefault(name, {})

        def set(self, key: str, value: StorageBase.VALUE_TYPE):
            """Set the given key/value in the given namespace"""
            self._data[key] = value

        def get(self, key: str) -> StorageBase.VALUE_TYPE:
            """Get the value for the given key in the given namespace"""
            return self._data.get(key)

        def pop(self, key: str) -> StorageBase.VALUE_TYPE:
            """Delete the key from the namespace and return any value that was
            set
            """
            return self._data.pop(key, None)

    def namespace(self, name: str) -> DictStorageNamespace:
        return self.DictStorageNamespace(name, self)



"""
Extensible set of ingestors that can link to multiple RAG consumers
"""

# Local
from ..factory import ImportableFactory
from .anything_llm import AnythingLLMIngestor
from .open_webui import OpenWebUIIngestor

ingestor_factory = ImportableFactory("ingestor")
ingestor_factory.register(AnythingLLMIngestor)
ingestor_factory.register(OpenWebUIIngestor)



"""
This toolkit utility provides common factory construction semantics and a common
base class for classes that can be constructed via config

Derived from: https://github.com/caikit/caikit/blob/main/caikit/core/toolkit/factory.py
"""

# Standard
import abc
import copy
import importlib

# Third Party
import jsonschema
import jsonschema.exceptions

# First Party
import aconfig
import alog

# Local
from .config.merge import merge_configs

log = alog.use_channel("FCTRY")


class FactoryConstructible(abc.ABC):
    """A class can be constructed by a factory if its constructor takes exactly
    one argument that is an aconfig.Config object and it has a name to identify
    itself with the factory.
    """

    # This is the set of default values for the config that me be set by child
    # implementations
    config_defaults = {}

    @property
    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """This is the name of this constructible type that will be used by
        the factory to identify this class
        """

    @property
    @classmethod
    @abc.abstractmethod
    def config_schema(cls) -> dict:
        """This is the jsonschema dict for this instance's config"""

    @abc.abstractmethod
    def __init__(self, config: aconfig.Config, instance_name: str, **kwargs):
        """A FactoryConstructible object must be constructed with a config
        object that it uses to pull in all configuration
        """


class Factory:
    """The base factory class implements all common factory functionality to
    read a designated portion of config and instantiate an instance of the
    registered classes.
    """

    # The keys in the instance config
    TYPE_KEY = "type"
    CONFIG_KEY = "config"

    def __init__(self, name: str):
        """Construct with the path in the global config where this factory's
        configuration lives.
        """
        self._name = name
        self._registered_types = {}

    @property
    def name(self) -> str:
        return self._name

    def register(self, constructible: type[FactoryConstructible]):
        """Register the given constructible"""
        current = self._registered_types.get(constructible.name)
        assert (
            current is None or current is constructible
        ), f"Conflicting registration of {constructible.name}"
        self._registered_types[constructible.name] = constructible

    def registered_types(self) -> list[str]:
        """Get the list of registered types"""
        return list(sorted(self._registered_types.keys()))

    def construct(
        self,
        instance_config: dict,
        instance_name: str | None = None,
        *,
        validate: bool = True,
        **kwargs,
    ) -> FactoryConstructible:
        """Construct an instance of the given type"""
        inst_type = instance_config.get(self.__class__.TYPE_KEY)
        inst_cls = self._registered_types.get(inst_type)
        inst_cfg = aconfig.Config(
            merge_configs(
                copy.deepcopy(inst_cls.config_defaults),
                instance_config.get(self.__class__.CONFIG_KEY, {}),
            ),
            override_env_vars=False,
        )
        assert (
            inst_cls is not None
        ), f"No {self.name} class registered for type {inst_type}"
        instance_name = instance_name or inst_cls.name
        if validate:
            try:
                # NOTE: This explicitly allows additional properties
                jsonschema.validate(instance=inst_cfg, schema=inst_cls.config_schema)
            except jsonschema.exceptions.ValidationError as err:
                log.error(
                    "Failed to construct [%s] due to invalid config: %s",
                    instance_name,
                    err,
                )
                raise
        return inst_cls(inst_cfg, instance_name, **kwargs)


class ImportableFactory(Factory):
    """An ImportableFactory extends the base Factory to allow the construction
    to specify an "import_class" field that will be used to import and register
    the implementation class before attempting to initialize it.
    """

    IMPORT_CLASS_KEY = "import_class"

    def construct(
        self,
        instance_config: dict,
        instance_name: str | None = None,
        **kwargs,
    ):
        # Look for an import_class and import and register it if found
        import_class_val = instance_config.get(self.__class__.IMPORT_CLASS_KEY)
        if import_class_val:
            assert isinstance(str, import_class_val)
            module_name, class_name = import_class_val.rsplit(".", 1)
            imported_module = importlib.import_module(module_name)
            imported_class = getattr(imported_module, class_name)
            assert issubclass(imported_class, FactoryConstructible)

            self.register(imported_class)
        return super().construct(instance_config, instance_name, **kwargs)



"""
Module to manage merging user config with base config

Based on https://github.com/caikit/caikit/blob/main/caikit/config/config.py
"""
# Standard
from enum import Enum

# First Party
import aconfig

_CONFIG_TYPE = dict | aconfig.Config


class ListMergeType(Enum):
    REPLACE = "replace"
    PREPEND = "prepend"
    APPEND = "append"
    INDEX = "index"


def merge_configs(
    base: _CONFIG_TYPE | None,
    overrides: _CONFIG_TYPE | None,
    list_merge_type: ListMergeType = ListMergeType.REPLACE,
) -> aconfig.Config:
    """Helper to perform a deep merge of the overrides into the base. The merge
    is done in place, but the resulting dict is also returned for convenience.
    The merge logic is quite simple: If both the base and overrides have a key
    and the type of the key for both is a dict, recursively merge, otherwise
    set the base value to the override value.
    Args:
        base: The base config that will be updated with the overrides
        overrides: The override config
        list_merge_type: The type of merge logic to use for lists
    Returns:
        merged:
            The merged results of overrides merged onto base
    """
    # Handle none args
    if base is None:
        return overrides or {}
    if overrides is None:
        return base or {}

    # Do the deep merge
    for key, value in overrides.items():
        if (
            key not in base
            or not isinstance(base[key], (dict, list))
            or not isinstance(value, (dict, list))
        ):
            base[key] = value
        elif isinstance(value, list):
            if list_merge_type == ListMergeType.REPLACE:
                base[key] = value
            elif list_merge_type == ListMergeType.PREPEND:
                base[key] = value + base[key]
            elif list_merge_type == ListMergeType.APPEND:
                base[key] = base[key] + value
            elif list_merge_type == ListMergeType.INDEX:
                orig_list = base[key]
                for i, entry in enumerate(value):
                    if i < len(orig_list):
                        orig_list[i] = entry
                    else:
                        orig_list.append(entry)
            else:
                raise ConfigError(f"Bad list merge type: {list_merge_type}")
        else:
            base[key] = merge_configs(base[key], value)

    return base



"""
Config initializer for AnythingLLM
"""
# Standard
import os
import platform

# Local
from ...ingestors import AnythingLLMIngestor
from ...ingestors.base import Ingestor
from .initializer_base import IngestorInitializerBase


class AnythingLLMInitializer(IngestorInitializerBase):
    __doc__ = __doc__

    def ingestor_class(self) -> type[Ingestor]:
        return AnythingLLMIngestor

    def is_installed(self):
        system = platform.system()
        if system == "Darwin":
            return os.path.exists("/Applications/AnythingLLM.app")
        print(f"Auto-detection not supported on {system} for AnythingLLM")
        return False



"""
Base class for a self-contained command
"""

# Standard
import abc
import argparse


class CommandBase(abc.ABC):
    __doc__ = __doc__

    @property
    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """The name of the command that will be used to select it"""

    @abc.abstractmethod
    def add_args(self, parser: argparse.ArgumentParser):
        """Add the arguments for this command"""

    @abc.abstractmethod
    def run(self, args: argparse.Namespace):
        """Run the command with the given parsed"""
